Imports System.ComponentModel
Imports System.Runtime.InteropServices
Imports System.Windows.Forms.VisualStyles
Imports ThemedRichTextBox.NativeMethods
Imports WinForms = System.Windows.Forms

''' <summary>
'''   Represents a richtextbox control.  This class extends the Windows Forms
'''   richtextbox control by support for visual styles.
''' </summary>
Public Class ExtendedRichTextBox
    Inherits RichTextBox

    Private m_BorderPen As Pen
    Private m_BackColorPen As Pen

    ''' <summary>
    '''   Initializes a new instance of the <c>RichTextBox</c> class.
    ''' </summary>
    Public Sub New()
        m_BorderPen = New Pen(VisualStyleInformation.TextControlBorder)
        m_BackColorPen = New Pen(Me.BackColor)
        Me.SetStyle(ControlStyles.ResizeRedraw, True)
        Me.UpdateStyles()
    End Sub

    ''' <summary>
    '''   Gets a value indicating whether the horizontal scrollbar is visible on
    '''   the control.
    ''' </summary>
    ''' <value><c>True</c>, if the horizontal scrollbar is visible.</value>
    Public ReadOnly Property HorizontalScrollBarVisible() As Boolean
        Get
            Dim sbi As New SCROLLBARINFO()
            sbi.cbSize = Marshal.SizeOf(sbi)
            If Not GetScrollBarInfo(Me.Handle, CONST_OBJID_HSCROLL, sbi) Then
                Throw New Win32Exception()
            End If
            Return (sbi.rgstate(0) <> STATE_SYSTEM_INVISIBLE)
        End Get
    End Property

    ''' <summary>
    '''   Gets a value indicating whether the vertical scrollbar is visible on
    '''   the control.
    ''' </summary>
    ''' <value><c>True</c>, if the vertical scrollbar is visible.</value>
    Public ReadOnly Property VerticalScrollBarVisible() As Boolean
        Get
            Dim sbi As New SCROLLBARINFO()
            sbi.cbSize = Marshal.SizeOf(sbi)
            If Not GetScrollBarInfo(Me.Handle, CONST_OBJID_VSCROLL, sbi) Then
                Throw New Win32Exception()
            End If
            Return (sbi.rgstate(0) <> STATE_SYSTEM_INVISIBLE)
        End Get
    End Property

    Protected Overrides Sub OnResize(ByVal e As EventArgs)
        Me.RedrawFrame()
        MyBase.OnResize(e)
    End Sub

    Protected Overrides Sub OnBackColorChanged(ByVal e As EventArgs)
        If m_BackColorPen IsNot Nothing Then
            m_BackColorPen.Dispose()
        End If
        m_BackColorPen = New Pen(Me.BackColor)

        Me.RedrawFrame()
        MyBase.OnBackColorChanged(e)
    End Sub

    Protected Overrides Sub OnEnabledChanged(ByVal e As EventArgs)
        Me.RedrawFrame()
        MyBase.OnEnabledChanged(e)
    End Sub

    Protected Overrides Sub WndProc(ByRef m As Message)
        If _
            Not Application.RenderWithVisualStyles OrElse _
            Me.BorderStyle <> WinForms.BorderStyle.Fixed3D _
        Then
            MyBase.WndProc(m)
            Return
        End If
        Select Case m.Msg

            ' Draw the control's nonclient area (border and scrollbars).
            Case WM_NCPAINT
                Dim hDC As IntPtr = GetWindowDC(Me.Handle)
                Using g As Graphics = Graphics.FromHdc(hDC)
                    Dim pos As Point = Me.Parent.PointToScreen(Me.Location)

                    ' Update occurs only inside the region.  Thus Windows does
                    ' not draw the window's border.
                    Using _
                        r As New Region( _
                            New Rectangle( _
                                pos.X + 2, _
                                pos.Y + 2, _
                                Me.Width - 3, _
                                Me.Height - 3 _
                            ) _
                        )

                        m.WParam = r.GetHrgn(g)

                        ' Let Windows draw the control's scrollbars.
                        MyBase.WndProc(m)

                        r.ReleaseHrgn(m.WParam)
                    End Using

                    ' Fill the rectangle between the scrollbars if the
                    ' scrollbars are both visible:
                    '
                    '        | |
                    '        | |
                    '   -----+-+
                    '   -----+-+
                    If _
                        Me.HorizontalScrollBarVisible AndAlso _
                        Me.VerticalScrollBarVisible _
                    Then
                        Dim Left As Integer = 2
                        If Me.RightToLeft <> WinForms.RightToLeft.Yes Then
                            Left += Me.ClientSize.Width
                        End If
                        g.FillRectangle( _
                            SystemBrushes.Control, _
                            Left, _
                            Me.Height - 2 - _
                                SystemInformation.HorizontalScrollBarHeight, _
                            SystemInformation.VerticalScrollBarWidth, _
                            SystemInformation.HorizontalScrollBarHeight _
                        )
                    End If
                    Dim BackColorPen As Pen
                    If Me.Enabled Then
                        BackColorPen = m_BackColorPen
                    Else
                        BackColorPen = SystemPens.Control
                    End If
                    g.DrawRectangle( _
                        BackColorPen, _
                        1, 1, _
                        Me.Width - 3, Me.Height - 3 _
                    )
                    g.DrawRectangle( _
                        m_BorderPen, _
                        0, 0, _
                        Me.Width - 1, Me.Height - 1 _
                    )
                End Using
                ReleaseDC(Me.Handle, hDC)

                ' Message has been processed.
                m.Result = IntPtr.Zero

                ' Set location and size of the control's nonclient area.
            Case WM_NCCALCSIZE
                If _
                    m.WParam = IntPtr.Zero OrElse _
                    m.WParam = New IntPtr(1) _
                Then
                    Dim nccsp As NCCALCSIZE_PARAMS = _
                        DirectCast( _
                            m.GetLParam(GetType(NCCALCSIZE_PARAMS)), _
                            NCCALCSIZE_PARAMS _
                        )
                    Dim rct As RECT = nccsp.rgrc0
                    rct.Top += 2
                    rct.Left += 2
                    rct.Right -= 2
                    rct.Bottom -= 2
                    If Me.VerticalScrollBarVisible Then
                        If Me.RightToLeft = WinForms.RightToLeft.Yes Then
                            rct.Left += _
                                SystemInformation.VerticalScrollBarWidth
                        Else
                            rct.Right -= _
                                SystemInformation.VerticalScrollBarWidth
                        End If
                    End If
                    If Me.HorizontalScrollBarVisible Then
                        rct.Bottom -= _
                            SystemInformation.HorizontalScrollBarHeight
                    End If
                    nccsp.rgrc0 = rct

                    ' Copy back the new client rectangle.
                    Marshal.StructureToPtr(nccsp, m.LParam, False)
                End If

                ' Update the pen used to draw the themed border.
            Case WM_THEMECHANGED
                If m_BorderPen IsNot Nothing Then
                    m_BorderPen.Dispose()
                End If
                m_BorderPen = _
                    New Pen(VisualStyleInformation.TextControlBorder)
                m.Result = IntPtr.Zero
                MyBase.WndProc(m)
            Case Else
                MyBase.WndProc(m)
        End Select
    End Sub

    Private Sub RedrawFrame()

        ' Redraw the window's frame.
        Const RedrawFlags As RDW_Flags = _
            RDW_Flags.RDW_INVALIDATE Or _
            RDW_Flags.RDW_FRAME Or _
            RDW_Flags.RDW_NOERASE Or _
            RDW_Flags.RDW_NOCHILDREN Or _
            RDW_Flags.RDW_UPDATENOW
        If _
            Not RedrawWindow( _
                Me.Handle, _
                New RECT(0, 0, Me.Height, Me.Width), _
                IntPtr.Zero, _
                RedrawFlags _
            ) _
        Then
            Throw New Win32Exception()
        End If
    End Sub
End Class